/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  cachemgr.c -- Return cache files, and implement old cache file removal.

  **************************************/

#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/signal.h>
#include <sys/un.h>
#include <string.h>
#include <utime.h>
#include <dirent.h>
#include <stdio.h>
#include <fcntl.h>

#include "common.h"

/* ------------------------------------------------------------- **
** This is the cache manager bit. Each file in the cache has a
** cache_entry structure which is held in a hash table. The end of the
** filename should be well distributed and is used as the hashing key. 
** Each cache_entry structure also has a pair of pointers, and so can
** also be accessed as a doubly linked list with most recently
** accessed files nearer the head.
** 
** A running tally is kept of total file size, and the LRU cache file
** removed whenever this exceeds the maximum. 
**
** When the client wishes to retrieve a URL it sends a string of the
** form "G URL mdtm size type offset\n". If the file is cached then
** the reply is a "R" with the file descriptor to read to attatched to
** the message. If the file isn't cached the reply is a "W" with the
** file descriptor to write to. A "\0" is sent on error.
** ------------------------------------------------------------- */

struct cache_entry {
	struct cache_entry *next, *prev, *collision;

	time_t last_access;
	sstr *filename;
	sstr *uri;
	sstr *mdtm;
	int size, cached;
};

sstr *cachemgr_init(char *cd, int cs);

static void scan_dir(const char *dir);
static int init_dir(const char *dir);
static struct cache_entry *make_entry(char *name);
static struct cache_entry **hash_loc(sstr *key);
static void add_entry(struct cache_entry *entry);
static void purge_entry(struct cache_entry *entry);
static void remove_entry(struct cache_entry *entry);
static void free_entry(struct cache_entry *p);
static void lru_entry(struct cache_entry *p);
void cmgr_run(int fd);
int uri_request(sstr *req, int *retfd);
int open_cachefile(struct cache_entry *p, int offset);
int make_cachefile(struct cache_entry *p);
sstr *cached_filename(const sstr *uri);

#define table_size 256

static struct cache_entry *hash_table[table_size];
static struct cache_entry *head = NULL, *tail = NULL;
static unsigned long total_size = 0;
static int cachesize;
static sstr *cachedir;

sstr *cachemgr_init(char *cd, int cs)
{
	int fd;
	struct sockaddr_un listen_addr;
	sstr *socket_file;

	if (init_dir(cd) == -1)
		return (NULL);

	cachedir = sstr_init(0);
	sstr_apprintf(cachedir, "%s/cache/", cd);
	cachesize = cs;
	socket_file = sstr_dup(cachedir);
	sstr_apprintf(socket_file, "/frox-cache");

	unlink(sstr_buf(socket_file));
	listen_addr.sun_family=AF_UNIX;
	strcpy(listen_addr.sun_path, sstr_buf(socket_file));
	
	if( (fd = socket(PF_UNIX, SOCK_STREAM, 0))==-1) {
		debug_perr("socket");
		return NULL;
	}

	if(bind(fd, (struct sockaddr*) &listen_addr, sizeof(listen_addr))==-1){
		debug_perr("bind");
		fprintf(stderr, "Check frox has permission to create a" 
			" file in %s\n", cd);
		return NULL;
	}
	
	if(listen(fd, 10)==-1) {
		debug_perr("bind");
		return NULL;
	}

	switch (fork()) {
	case -1:
		return (NULL);
	case 0:
		signal(SIGHUP, SIG_IGN);
		break;
	default:
		close(fd);
		return (socket_file);
	}

	sstr_free(socket_file);
	
	scan_dir(sstr_buf(cachedir));

	cmgr_run(fd);
	exit(-1);
}

void cmgr_run(int listen)
{
	fd_set reads;
	sstr *buf;
	int cfd, fd;
	buf=sstr_init(MAX_LINE_LEN);

	do {
		fd = accept(listen, NULL, 0);
		if(fd==-1) {
			debug_perr("accept");
			continue;
		}

		FD_ZERO(&reads);
		FD_SET(fd, &reads);
		select(fd+1, &reads, NULL, NULL, NULL);
		sstr_empty(buf);
		sstr_append_read(fd, buf, MAX_LINE_LEN * 2 - 1);
		
		switch(sstr_getchar(buf, 0)){
		case 'G':
			switch(uri_request(buf, &cfd)){
			case -1:
				write(fd, "\0", 1);
				break;
			case 1:
				send_fd(fd, cfd, 'R');
				close(cfd);
				break;
			case 0:
				send_fd(fd, cfd, 'W');
				close(cfd);
				break;
			}
			break;
		default:
			debug_err("Arrgh - unknown cachemgr cmd");
			debug2("%s\n", sstr_buf(buf));
			write(fd, "\0", 1);
			close(fd);
		}
		close(fd);
	} while(TRUE);
}

int uri_request(sstr *req, int *retfd)
{
	struct cache_entry *p;
	static sstr *uri=NULL, *mdtm=NULL;
	sstr *filename;
	int size, offset, type;

	if(!uri) uri=sstr_init(MAX_LINE_LEN);
	if(!mdtm) mdtm=sstr_init(MAX_LINE_LEN);

	sstr_token(req, NULL, " ", 0);
	sstr_token(req, uri, " ", 0);
	sstr_token(req, mdtm, " ", 0);
	size = sstr_atoi(req);
	sstr_token(req, NULL, " ", 0);
	type = sstr_atoi(req);
	sstr_token(req, NULL, " ", 0);
	offset = sstr_atoi(req);
			
	filename = cached_filename(uri);

	p=*hash_loc(filename);
	if(p) {
		if(!sstr_cmp(uri, p->uri) && !sstr_cmp(mdtm, p->mdtm) && 
		   p->size==size){
			debug("Cache hit\n");
			*retfd=open_cachefile(p, offset);
			if(*retfd != -1) {
				if(p->cached==p->size) return 1;
				else return 0;
			} else {
				debug("Offset > cached bytes. Can't cache\n");
				return -1;
			}
		}
		debug("Invalid cache file\n");
		purge_entry(p);
	}

	if(offset) { /*Can't start caching new file from halfway through*/
		debug("Only part of file requested - not caching\n");
		return -1;
	}
	debug("Not cached\n");	
	p = (struct cache_entry *) malloc(sizeof(struct cache_entry));
	p->uri = sstr_dup(uri);
	p->filename = sstr_dup(filename);
	p->mdtm = sstr_dup(mdtm);
	p->size=size;
	p->cached=0;
	p->last_access=time(NULL);
	*retfd=make_cachefile(p);
	if(*retfd==-1) {
		write_log(ERROR, "Creating cache file %s failed."
			  "Stopping caching", filename);
		free_entry(p);
		return(-1);
	}

	add_entry(p);
	return 0;
}

int make_cachefile(struct cache_entry *p)
{
	int ret;
	sstr *buf;

	ret = creat(sstr_buf(p->filename), S_IRUSR | S_IWUSR);
	if (ret == -1) return (-1);

	buf=sstr_init(MAX_LINE_LEN*2);
	sstr_apprintf(buf, "%.3d  %s %.12d %.1d %s %.12lu\n", 
		      sstr_len(p->mdtm) + sstr_len(p->uri) + 30,
		      sstr_buf(p->mdtm), p->size, 1, sstr_buf(p->uri), 
		      p->last_access);
	sstr_write(ret, buf, 0);
	sstr_free(buf);
	return ret;
}			  

int open_cachefile(struct cache_entry *p, int offset)
{
	int fd, i;
	struct stat status;
	char buf2[13];
	sstr *buf;

	debug("Opening cache file\n");
	fd = open(sstr_buf(p->filename), O_RDWR);
	buf=sstr_init(0);

	lseek(fd, 0, SEEK_SET);
	sstr_append_read(fd, buf, 5);
	i=sstr_atoi(buf);

	fstat(fd, &status);
	p->cached = status.st_size - i - 5;
	if(offset > p->cached) {
		sstr_free(buf);
		return -1;
	}

	sstr_append_read(fd, buf, i);
	lru_entry(p);
	p->last_access=time(NULL);
	lseek(fd, sstr_len(buf)-13, SEEK_SET);
	sprintf(buf2, "%.12lu", p->last_access);
	write(fd, buf2, 12);
	lseek(fd, sstr_len(buf), SEEK_SET);
	lseek(fd, offset, SEEK_CUR);
	sstr_free(buf);
	return fd;
}

static void lru_entry(struct cache_entry *p)
{
	if(head == p) return;
	if(!p->prev) debug_err("Linked list screwed up");

	if (p->next) p->next->prev = p->prev;
	else tail = p->prev;

	p->prev->next = p->next;
	
	p->next = head;
	head->prev = p;
	p->prev = NULL;
	head = p;
}

struct cache_entry **hash_loc(sstr *key)
{
	int i;
	struct cache_entry *p;

	i = strtol(sstr_buf(key) + sstr_len(key) - 4, NULL, 16) % table_size;

	if (hash_table[i] == NULL)
		return (&hash_table[i]);
	if (!sstr_cmp(hash_table[i]->filename, key))
		return (&hash_table[i]);

	debug("X");		/*Collision */
	for (p = hash_table[i];
	     p->collision != NULL && sstr_cmp(p->collision->filename, key);
	     p = p->collision);
	return (&p->collision);
}

void add_entry(struct cache_entry *entry)
{
	struct cache_entry *p, **pp;

	entry->collision = NULL;

	pp = hash_loc(entry->filename);

	if (*pp != NULL) {
		debug("   Already got it - removing old entry.\n");

		remove_entry(*pp);
		pp = hash_loc(entry->filename);
	}
	*pp = entry;

	/*Now add to linked list, based on last access time */
	for (p = head; p != NULL && entry->last_access < p->last_access;
	     p = p->next);
	entry->next = p;
	entry->prev = p ? p->prev : tail;
	if (entry->prev)
		entry->prev->next = entry;
	else
		head = entry;
	if (entry->next)
		entry->next->prev = entry;
	else
		tail = entry;

	total_size += entry->size;
	if (cachesize)
		while (total_size > cachesize)
			purge_entry(tail);
}

void purge_entry(struct cache_entry *entry)
{
	unlink(sstr_buf(entry->filename));
	remove_entry(entry);
}

void remove_entry(struct cache_entry *entry)
{
	int index;
	struct cache_entry *p;

	if (entry->next)
		entry->next->prev = entry->prev;
	else
		tail = entry->prev;
	if (entry->prev)
		entry->prev->next = entry->next;
	else
		head = entry->next;

	index = strtol(sstr_buf(entry->filename) + sstr_len(entry->filename)
		       - 4, NULL, 16) % table_size;

	debug2("Removing file %s\n", sstr_buf(entry->uri));
	if (hash_table[index] == NULL) {
		debug_err("Arrghhhhhhhhhhhhhhhhh");
		exit(-1);
	}

	if (entry == hash_table[index]) {
		hash_table[index] = entry->collision;
	} else {
		for (p = hash_table[index];
		     p->collision != NULL && p->collision != entry;
		     p = p->collision);
		if (p->collision == NULL) {
			debug("Removing non existant entry!\n");
			exit(-1);
		}
		p->collision = entry->collision;
	}

	total_size -= entry->size;
	free_entry(entry);
}

static void free_entry(struct cache_entry *p)
{
	sstr_free(p->filename);
	sstr_free(p->uri);
	sstr_free(p->mdtm);
	free(p);
}

/* ------------------------------------------------------------- **
** The following is an MD5sum algorithm brutally hacked from Colin
** Plumb's public domain code. 
** ------------------------------------------------------------- */
#define UWORD32 u_int32_t
void MD5Transform(UWORD32 buf[4], UWORD32 const in[16]);

#ifdef WORDS_BIGENDIAN
void byteSwap(UWORD32 * buf, unsigned words)
{
	unsigned char *p = (unsigned char *) buf;

	do {
		*buf++ = (UWORD32) ((unsigned) p[3] << 8 | p[2]) << 16 |
		    ((unsigned) p[1] << 8 | p[0]);
		p += 4;
	} while (--words);
}
#else
#define byteSwap(buf,words)
#endif

/* ------------------------------------------------------------- **
** Hash uri and return the filename (with absolute path) that this
** file should be cached in. 
** ------------------------------------------------------------- */
sstr *cached_filename(const sstr *u)
{
	static sstr *filename=NULL;
	UWORD32 buf[4];
	UWORD32 bytes[2];
	UWORD32 in[16];
	unsigned char *p;
	int len, i;
	const char *uri = sstr_buf(u);

	if(filename==NULL) filename=sstr_init(MAX_LINE_LEN);

	sstr_cpy(filename, cachedir);

	buf[0] = 0x67452301;
	buf[1] = 0xefcdab89;
	buf[2] = 0x98badcfe;
	buf[3] = 0x10325476;

	bytes[0] = 0;
	bytes[1] = 0;

	len = sstr_len(u);
	bytes[0] = len;

	/* Process data in 64-byte chunks */
	while (len >= 64) {
		memcpy(in, uri, 64);
		byteSwap(in, 16);
		MD5Transform(buf, in);
		uri += 64;
		len -= 64;
	}

	/* Handle any remaining bytes of data. */
	memcpy(in, uri, len);

	p = (unsigned char *) in + len;

	/* Set the first char of padding to 0x80.  There is always room. */
	*p++ = 0x80;

	/* Bytes of padding needed to make 56 bytes (-8..55) */
	len = 56 - 1 - len;

	if (len < 0) {		/* Padding forces an extra block */
		memset(p, 0, len + 8);
		byteSwap(in, 16);
		MD5Transform(buf, in);
		p = (unsigned char *) in;
		len = 56;
	}
	memset(p, 0, len);
	byteSwap(in, 14);

	/* Append length in bits and transform */
	in[14] = bytes[0] << 3;
	in[15] = bytes[1] << 3 | bytes[0] >> 29;
	MD5Transform(buf, in);

	byteSwap(buf, 4);

	p = (unsigned char *) buf;
	sstr_apprintf(filename, "%02x/", (*p++) % 0x0F);
	for (i = 0; i < 15; ++i) sstr_apprintf(filename, "%02x", *p++);

	return (filename);
}

/* The four core functions - F1 is optimized somewhat */

/* #define F1(x, y, z) (x & y | ~x & z) */
#define F1(x, y, z) (z ^ (x & (y ^ z)))
#define F2(x, y, z) F1(z, x, y)
#define F3(x, y, z) (x ^ y ^ z)
#define F4(x, y, z) (y ^ (x | ~z))

/* This is the central step in the MD5 algorithm. */
#define MD5STEP(f,w,x,y,z,in,s) \
	 (w += f(x,y,z) + in, w = (w<<s | w>>(32-s)) + x)

/*
 * The core of the MD5 algorithm, this alters an existing MD5 hash to
 * reflect the addition of 16 longwords of new data.  MD5Update blocks
 * the data and converts bytes into longwords for this routine.
 */
void MD5Transform(UWORD32 buf[4], UWORD32 const in[16])
{
	register UWORD32 a, b, c, d;

	a = buf[0];
	b = buf[1];
	c = buf[2];
	d = buf[3];

	MD5STEP(F1, a, b, c, d, in[0] + 0xd76aa478, 7);
	MD5STEP(F1, d, a, b, c, in[1] + 0xe8c7b756, 12);
	MD5STEP(F1, c, d, a, b, in[2] + 0x242070db, 17);
	MD5STEP(F1, b, c, d, a, in[3] + 0xc1bdceee, 22);
	MD5STEP(F1, a, b, c, d, in[4] + 0xf57c0faf, 7);
	MD5STEP(F1, d, a, b, c, in[5] + 0x4787c62a, 12);
	MD5STEP(F1, c, d, a, b, in[6] + 0xa8304613, 17);
	MD5STEP(F1, b, c, d, a, in[7] + 0xfd469501, 22);
	MD5STEP(F1, a, b, c, d, in[8] + 0x698098d8, 7);
	MD5STEP(F1, d, a, b, c, in[9] + 0x8b44f7af, 12);
	MD5STEP(F1, c, d, a, b, in[10] + 0xffff5bb1, 17);
	MD5STEP(F1, b, c, d, a, in[11] + 0x895cd7be, 22);
	MD5STEP(F1, a, b, c, d, in[12] + 0x6b901122, 7);
	MD5STEP(F1, d, a, b, c, in[13] + 0xfd987193, 12);
	MD5STEP(F1, c, d, a, b, in[14] + 0xa679438e, 17);
	MD5STEP(F1, b, c, d, a, in[15] + 0x49b40821, 22);

	MD5STEP(F2, a, b, c, d, in[1] + 0xf61e2562, 5);
	MD5STEP(F2, d, a, b, c, in[6] + 0xc040b340, 9);
	MD5STEP(F2, c, d, a, b, in[11] + 0x265e5a51, 14);
	MD5STEP(F2, b, c, d, a, in[0] + 0xe9b6c7aa, 20);
	MD5STEP(F2, a, b, c, d, in[5] + 0xd62f105d, 5);
	MD5STEP(F2, d, a, b, c, in[10] + 0x02441453, 9);
	MD5STEP(F2, c, d, a, b, in[15] + 0xd8a1e681, 14);
	MD5STEP(F2, b, c, d, a, in[4] + 0xe7d3fbc8, 20);
	MD5STEP(F2, a, b, c, d, in[9] + 0x21e1cde6, 5);
	MD5STEP(F2, d, a, b, c, in[14] + 0xc33707d6, 9);
	MD5STEP(F2, c, d, a, b, in[3] + 0xf4d50d87, 14);
	MD5STEP(F2, b, c, d, a, in[8] + 0x455a14ed, 20);
	MD5STEP(F2, a, b, c, d, in[13] + 0xa9e3e905, 5);
	MD5STEP(F2, d, a, b, c, in[2] + 0xfcefa3f8, 9);
	MD5STEP(F2, c, d, a, b, in[7] + 0x676f02d9, 14);
	MD5STEP(F2, b, c, d, a, in[12] + 0x8d2a4c8a, 20);

	MD5STEP(F3, a, b, c, d, in[5] + 0xfffa3942, 4);
	MD5STEP(F3, d, a, b, c, in[8] + 0x8771f681, 11);
	MD5STEP(F3, c, d, a, b, in[11] + 0x6d9d6122, 16);
	MD5STEP(F3, b, c, d, a, in[14] + 0xfde5380c, 23);
	MD5STEP(F3, a, b, c, d, in[1] + 0xa4beea44, 4);
	MD5STEP(F3, d, a, b, c, in[4] + 0x4bdecfa9, 11);
	MD5STEP(F3, c, d, a, b, in[7] + 0xf6bb4b60, 16);
	MD5STEP(F3, b, c, d, a, in[10] + 0xbebfbc70, 23);
	MD5STEP(F3, a, b, c, d, in[13] + 0x289b7ec6, 4);
	MD5STEP(F3, d, a, b, c, in[0] + 0xeaa127fa, 11);
	MD5STEP(F3, c, d, a, b, in[3] + 0xd4ef3085, 16);
	MD5STEP(F3, b, c, d, a, in[6] + 0x04881d05, 23);
	MD5STEP(F3, a, b, c, d, in[9] + 0xd9d4d039, 4);
	MD5STEP(F3, d, a, b, c, in[12] + 0xe6db99e5, 11);
	MD5STEP(F3, c, d, a, b, in[15] + 0x1fa27cf8, 16);
	MD5STEP(F3, b, c, d, a, in[2] + 0xc4ac5665, 23);

	MD5STEP(F4, a, b, c, d, in[0] + 0xf4292244, 6);
	MD5STEP(F4, d, a, b, c, in[7] + 0x432aff97, 10);
	MD5STEP(F4, c, d, a, b, in[14] + 0xab9423a7, 15);
	MD5STEP(F4, b, c, d, a, in[5] + 0xfc93a039, 21);
	MD5STEP(F4, a, b, c, d, in[12] + 0x655b59c3, 6);
	MD5STEP(F4, d, a, b, c, in[3] + 0x8f0ccc92, 10);
	MD5STEP(F4, c, d, a, b, in[10] + 0xffeff47d, 15);
	MD5STEP(F4, b, c, d, a, in[1] + 0x85845dd1, 21);
	MD5STEP(F4, a, b, c, d, in[8] + 0x6fa87e4f, 6);
	MD5STEP(F4, d, a, b, c, in[15] + 0xfe2ce6e0, 10);
	MD5STEP(F4, c, d, a, b, in[6] + 0xa3014314, 15);
	MD5STEP(F4, b, c, d, a, in[13] + 0x4e0811a1, 21);
	MD5STEP(F4, a, b, c, d, in[4] + 0xf7537e82, 6);
	MD5STEP(F4, d, a, b, c, in[11] + 0xbd3af235, 10);
	MD5STEP(F4, c, d, a, b, in[2] + 0x2ad7d2bb, 15);
	MD5STEP(F4, b, c, d, a, in[9] + 0xeb86d391, 21);

	buf[0] += a;
	buf[1] += b;
	buf[2] += c;
	buf[3] += d;
}

/********************************************************************
 * Stuff for scanning cache dir on startup to build data structures.
 * Don't need to be so carefull about security here.
 ********************************************************************/

void scan_dir(const char *dir)
{
	DIR *dp;
	struct dirent *entry;
	struct stat status;
	char name[512];

	dp = opendir(dir);
	if (dp == NULL) {
		debug_perr("opendir");
		write_log(ERROR, "Failed to open cache dir %s", dir);
		exit(-1);
	}
	chdir(dir);
	while ((entry = readdir(dp)) != NULL) {
		if (entry->d_name[0] == '.')
			continue;
		stat(entry->d_name, &status);
		if (S_ISDIR(status.st_mode))
			scan_dir(entry->d_name);
		else if (S_ISREG(status.st_mode)) {
			struct cache_entry *tmp;

			getcwd(name, 300);
			strcat(name, "/");
			strcat(name, entry->d_name);
			tmp = make_entry(name);
			if (tmp)
				add_entry(tmp);
		}
	}
	chdir("..");
	closedir(dp);
}

struct cache_entry *make_entry(char *name)
{
	struct stat status;
	struct cache_entry *ret;
	char buf[1024];
	int fd, i;

	ret = (struct cache_entry *) malloc(sizeof(struct cache_entry));
	if (ret == NULL) {
		debug_perr("malloc");
		exit(-1);
	}

	stat(name, &status);

	ret->filename = sstr_dup2(name);

	fd = open(name, O_RDONLY);
	read(fd, buf, 5);
	i = atoi(buf);
	if (i > 0 && i < 4096 && read(fd, buf, i) == i) {
		ret->mdtm = sstr_dup2(strtok(buf, " "));
		ret->size = atoi(strtok(NULL, " "));
		strtok(NULL, " ");	/*Type */
		ret->uri = sstr_dup2(strtok(NULL, " \r\n"));
		ret->last_access=strtoul(strtok(NULL, " \r\n"), NULL, 10);
		close(fd);
	} else {
		close(fd);
		debug2("Invalid file %s\n", name);
		unlink(name);
		sstr_free(ret->filename);
		free(ret);
		return (NULL);
	}

	ret->cached = status.st_size - i - 5;
	if (ret->size != ret->cached) {
		debug2("Partial file %s\n", name);
	}

	return (ret);
}

int init_dir(const char *dir)
{
	int i;
	struct stat status;
	char name[512];

	sprintf(name, "%s/cache", dir);
	if(stat(name, &status)==-1) {
		if (mkdir(name, S_IRWXU) == -1) {
			write_log(ERROR, "Unable to make cache dir %s",
				  name);
			return (-1);
		}
	}
	for (i = 0; i < 16; i++) {
		sprintf(name, "%s/cache/%02x", dir, i);
		if (stat(name, &status) == -1) {
			write_log(IMPORT, "Making cache dir %s", name);
			if (mkdir(name, S_IRWXU) == -1) {
				write_log(ERROR, "Unable to make cache dir %s",
					  name);
				return (-1);
			}
		} else if (!S_ISDIR(status.st_mode)) {
			write_log(ERROR, "%s is not a directory", name);
			return (-1);
		}
	}
	return (0);
}

